"""
    Module with invoice related entities
"""

from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import ( 
    T, 
    Tax,
    Item,
    Refund,
    Discount,
    DateRange,
    ActionLink, 
    PaypalName,
    AmountRange, 
    PayPalEntity, 
    ResponseType, 
    Money, 
    RecipientInfo,
    PaypalPhoneDetail,
    PaypalPortableAddress
)

from pypaypal.entities.invoicing.template import Template

class AggregatedDiscount(PayPalEntity):
    """Aggregated discount object representation
    """
    def __init__(self, invoice_discount: Discount, item_discount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.item_discount = item_discount
        self.invoice_discount = invoice_discount
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        itm_disc = Discount.serialize_from_json(json_data['item_discount'])
        inv_disc = Discount.serialize_from_json(json_data['invoice_discount'])

        return cls(inv_disc, itm_disc, json_response= json_data, response_type = response_type)

class ShippingCost(PayPalEntity):
    """Shipping Cost object representation
    """
    def __init__(self, tax: Tax, amount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.tax = tax
        self.amount = amount

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        tax = Discount.serialize_from_json(json_data['tax'])
        amt = Money.serialize_from_json(json_data['amount'])
        return cls(tax, amt, json_response= json_data, response_type = response_type)

class CustomAmount(PayPalEntity):
    """Custom amount object representation
    """
    def __init__(self, label: str, amount: Money, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.label = label
        self.amount = amount

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amt = Money.serialize_from_json(json_data['amount'])
        return cls(json_data['label'], amt, json_response= json_data, response_type = response_type)

class AmountWithBreakdown(PayPalEntity):
    """Amount with breakdown object representation
    """

    def __init__(self, item_total: Money, discount: AggregatedDiscount, tax_total: Money, shipping: ShippingCost, custom: CustomAmount, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))        
        self.item_total = item_total
        self.discount = discount
        self.tax_total = tax_total
        self.shipping = shipping
        self.custom = custom

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        item_total, discount, tax_total, shipping, custom = None, None, None, None, None
        
        if 'item_total' in json_data.keys():
            item_total = Money.serialize_from_json(json_data['item_total'])
        if 'discount' in json_data.keys():
            discount = AggregatedDiscount.serialize_from_json(json_data['discount'])
        if 'tax_total' in json_data.keys():
            tax_total = Money.serialize_from_json(json_data['tax_total'])
        if 'shipping' in json_data.keys():
            shipping = ShippingCost.serialize_from_json(json_data['shipping'])
        if 'custom' in json_data.keys():
            custom = CustomAmount.serialize_from_json(json_data['custom'])

        return cls(item_total, discount, tax_total, shipping, custom, json_response= json_data, response_type = response_type)

class InvoicePaymentTerm(PayPalEntity):
    """Invoice payment term object representation
    """

    def __init__(self, term_type: str, due_date: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self._due_date = due_date
        self.term_type = term_type
    
    @property
    def due_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._due_date) if self._due_date else None
        except:
            return None

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['due_date'] = ret.pop('_due_date', None)
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data['term_type'], json_data['due_date'], json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(cls, term_type: str, due_date: datetime):
        return cls(term_type, due_date.strftime('%Y-%m-%d'))

class FileReference(PayPalEntity):
    """File reference object representation
    """

    def __init__(self, file_id: str, reference_url: str, content_type: str, size: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = file_id
        self.reference_url = reference_url
        self.content_type = content_type
        self.size = size
        default_time = datetime.now().strftime('%Y-%m-%dT%H:%M:%s')
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time', default_time))

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['create_time'] = ret.pop('_create_time', None)
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('id'), json_data.get('reference_url'), json_data.get('content_type'),
            json_data.get('size'), json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def create(cls, file_id: str, reference_url: str, content_type: str, size: str, creation: datetime = datetime.now()) -> 'FileReference':
        return cls(file_id, reference_url, content_type, size, create_time = creation.strftime('%Y-%m-%dT%H:%M:%s'))

class MetaData(PayPalEntity):
    """Template audit metadata.
    """

    def __init__(self, created_by: str, last_updated_by: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.created_by = created_by
        self.last_updated_by = last_updated_by
        default_time = datetime.now().strftime('%Y-%m-%dT%H:%M:%s')
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time', default_time))
        self._last_update_time = self._json_response.get('last_update_time', kwargs.get('last_update_time', default_time))

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def last_updated_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._last_updated_time) if self._last_updated_time else None
        except:
            return None

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['create_time'] = ret.pop('_create_time', None)
        ret['last_updated_time'] = ret.pop('_last_updated_time', None)
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('created_by'), json_data.get('last_updated_by'), json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(cls, created_by: str, last_updated_by: str) -> 'MetaData':
        creation = datetime.now().strftime('%Y-%m-%dT%H:%M:%s')
        
        return cls(
            created_by, last_updated_by, 
            created_time = creation, last_update_time = creation
        )

class InvoiceDetail(PayPalEntity):
    """Invoice detail object representation
    """

    def __init__(
        self, currency_code: str, invoice_number: str, metadata: MetaData,
        payment_term: InvoicePaymentTerm=None, attachments: List[FileReference]= [],
         **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.metadata = metadata
        self.payment_term = payment_term
        self.currency_code = currency_code
        self.invoice_number = invoice_number
        self.attachments = attachments or []
        self.note = self._json_response.get('note', kwargs.get('note'))
        self.memo = self._json_response.get('memo', kwargs.get('memo'))
        self.reference = self._json_response.get('reference', kwargs.get('reference'))        
        self._invoice_date = self._json_response.get('invoice_date', kwargs.get('invoice_date'))
        self.terms_and_conditions = self._json_response.get('terms_and_conditions', kwargs.get('terms_and_conditions'))

    @property
    def invoice_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._invoice_date) if self._invoice_date else None
        except:
            return None

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['invoice_date'] = ret.pop('_invoice_date', None)
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        metadata, payment_term, attachments = None, None, []

        if 'metadata' in json_data.keys():
            metadata =  MetaData.serialize_from_json(json_data['metadata'], response_type)        
        if 'payment_term' in json_data.keys():
            payment_term = InvoicePaymentTerm.serialize_from_json(json_data['payment_term'], response_type)
        
        if 'attachments' in json_data.keys():
            attachments = [FileReference.serialize_from_json(x, response_type) for x in json_data['attachments']]

        return cls(
            json_data['currency_code'], json_data['invoice_number'], metadata, 
            payment_term, attachments, json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def create(
        cls, currency_code: str, invoice_number: str, metadata: MetaData, *, 
        payment_term: InvoicePaymentTerm=None, attachments: List[FileReference]= [],
        note:str = None, memo:str = None, reference:str = None, invoice_date: datetime = None, 
        terms_and_conditions: str = None
    ):
        return cls(
            currency_code, invoice_number, metadata, payment_term, 
            attachments or [], note = note, memo = memo, reference = reference, 
            invoice_date = invoice_date, terms_and_conditions = terms_and_conditions
        )

class InvoicerInfo(PayPalEntity):
    """Invoicer info object representation
    """

    def __init__(self, business_name: str, name: PaypalName = None, address : PaypalPortableAddress = None, phones: List[PaypalPhoneDetail] = None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.phones = phones
        self.address = address
        self.business_name = business_name
        self.tax_id = self._json_response.get('tax_id', kwargs.get('tax_id'))
        self.website = self._json_response.get('website', kwargs.get('website'))
        self.logo_url = self._json_response.get('logo_url', kwargs.get('logo_url'))
        self.additional_notes = self._json_response.get('additional_notes', kwargs.get('additional_notes'))

    def to_dict(self) -> dict:
        ret = super().to_dict()        
        if self.phones:
            ret['phones'] = [ x.to_dict() for x in self.phones ]
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        address, name, phones = None, None, []
        
        if 'name' in json_data.keys():
            name = PaypalName.serialize_from_json(json_data['name'], response_type)
        
        if 'address' in json_data.keys():
            address = PaypalPortableAddress.serialize_from_json(json_data['address'], response_type)

        if 'phones' in json_data.keys():
            phones = [PaypalPhoneDetail.serialize_from_json(x, response_type) for x in json_data['phones']]
        
        return cls(
            json_data.get('business_name'), name, address, phones, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(
        cls, business_name, *, name: PaypalName = None, 
        address : PaypalPortableAddress = None, phones: List[PaypalPhoneDetail] = None, 
        tax_id: str = None, website: str = None, logo_url: str = None, additional_notes: str = None 
    ):        
        return cls(
            business_name, name, address, phones, 
            tax_id = tax_id, website = website, logo_url = logo_url, 
            additional_notes = additional_notes
        )
    
class PartialPayment(PayPalEntity):
    """Partial payment object representation
    """
    
    def __init__(self, minimum_amount_due: Money, allow_partial_payment: bool, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.minimum_amount_due = minimum_amount_due
        self.allow_partial_payment = allow_partial_payment

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount = Money.serialize_from_json(json_data['minimum_amount_due'])
        return cls(amount, json_data.get('allow_partial_payment'), json_response= json_data, response_type = response_type)

    @classmethod
    def create(cls, minimum_amount_due: Money, allow_partial_payment: bool = False):
        return cls(minimum_amount_due, allow_partial_payment)

class InvoiceConfiguration(PayPalEntity):
    """Invoice configuration object representation
    """

    def __init__(self, template_id: str, tax_calc_after_discount: bool, tax_inclusive: bool, allow_tip: bool, partial_payment: PartialPayment=None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.allow_tip = allow_tip
        self.template_id = template_id
        self.tax_inclusive = tax_inclusive
        self.partial_payment = partial_payment
        self.tax_calc_after_discount = tax_calc_after_discount

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        partial_payment = PartialPayment.serialize_from_json(json_data['partial_payment']) if 'partial_payment' in json_data.keys() else None

        return cls(
            json_data.get('template_id'), json_data.get('tax_calc_after_discount'), json_data.get('tax_inclusive'),
            json_data.get('allow_tip'), partial_payment, json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def create(
        cls, *, template_id: str = None, tax_calc_after_discount: bool = True, 
        tax_inclusive: bool = False, allow_tip: bool = False, partial_payment: PartialPayment=None
    ):
        return cls(template_id, tax_calc_after_discount, tax_inclusive, allow_tip, partial_payment)

class AmountSummaryDetail(PayPalEntity):
    """Amount sumary object representation
    """

    def __init__(self, currency_code: str, value: str, breakdown: AmountWithBreakdown, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.value = value
        self.currency_code = currency_code
        self.breakdown = breakdown

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        breakdown = AmountWithBreakdown.serialize_from_json(json_data['breakdown'], response_type)
        return cls(json_data['currency_code'], json_data['value'], breakdown, json_response= json_data, response_type = response_type)
    
    @classmethod
    def create(cls, currency_code: str, value: str, breakdown: AmountWithBreakdown) -> 'AmountSummaryDetail':
        return cls(currency_code, value, breakdown)

class PaymentDetail(PayPalEntity):
    """Payment object representation for invoice draft
    """
    def __init__(self, pd_type: str, payment_id: str, method: str, amount: Money = None, shipping_info: ShippingCost= None, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.type = pd_type
        self.method = method
        self.amount = amount
        self.payment_id = payment_id
        self.shipping_info = shipping_info
        self.note = self._json_response.get('note', kwargs.get('note'))
        self._payment_date = self._json_response.get('payment_date', kwargs.get('payment_date'))

    @property
    def payment_date(self) -> datetime:
        try:
            return dateutil.parser.parse(self._payment_date) if self._payment_date else None
        except:
            return None

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['payment_date'] = ret.pop('_payment_date', None)
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount, shipping = None, None
        
        if 'amount' in json_data.keys():
            amount = Money.serialize_from_json(json_data['amount'])
        if 'shipping_info' in json_data.keys():
            shipping = ShippingCost.serialize_from_json(json_data['shipping_info'])

        return cls(json_data['pd_type'], json_data['payment_id'], json_data['method'], 
                    amount, shipping, json_response= json_data, response_type = response_type
            )
        
    @classmethod
    def create(cls, *, payment_id: str, method: str, note: str, payment_date: datetime, amount: Money, shipping_info: ShippingCost) -> T:
        return cls(None, payment_id, method, amount, shipping_info, note = note, payment_date = payment_date.strftime('%Y-%m-%d'))

class InvoicePayment(PayPalEntity):
    """Payment object representation for invoice draft
    """
    def __init__(self, paid_amount: Money, transactions: List[PaymentDetail], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.paid_amount = paid_amount
        self.transactions = transactions

    def to_dict(self) -> dict:
        ret = super().to_dict()
        if self.transactions:
            ret['transactions'] = [ x.to_dict() for x in self.transactions ]
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        paid_amount, transactions = None, list()

        if 'paid_amount' in json_data.keys():
            paid_amount = Money.serialize_from_json(json_data['paid_amount'], response_type)        
        if 'transactions' in json_data.keys():
            transactions = [PaymentDetail.serialize_from_json(x, response_type) for x in json_data['transactions']]
        
        return cls(paid_amount, transactions, json_response= json_data, response_type = response_type)

class Invoice(PayPalEntity):
    """Object representation for a paypal invoice
    """    
    def __init__(
        self, id:str, detail: InvoiceDetail, invoicer: InvoicerInfo, primary_recipients: List[RecipientInfo], 
        items: List[Item], configuration: InvoiceConfiguration, amount: Money, payments: List[InvoicePayment], 
        refunds: List[Refund], due_amount: Money = None, gratuity: Money = None, **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = id
        self.detail = detail
        self.invoicer = invoicer
        self.primary_recipients = primary_recipients
        self.items = items
        self.configuration = configuration
        self.amount = amount
        self.payments = payments
        self.refunds = refunds
        self.gratuity = gratuity
        self.due_amount = due_amount
        self.status = self._json_response.get('status', kwargs.get('status'))
        self.additional_recipients = self._json_response.get('additional_recipients', kwargs.get('additional_recipients'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def send_link(self) -> ActionLink:
        """Retrieves a link to send this draft and creating an invoice from it.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'send', self.links), None)

    @property
    def replace_link(self) -> ActionLink:
        """Retrieves a link to replace (update) this draft.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'replace', self.links), None)

    @property
    def delete_link(self) -> ActionLink:
        """Retrieves a link to delete this draft.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'delete', self.links), None)

    @property
    def record_payment_link(self) -> ActionLink:
        """Retrieves a link to record a payment on this draft.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'record-payment', self.links), None)

    @property
    def gen_qr_code_link(self) -> ActionLink:
        """Retrieves a link to generate a qr-code for this draft.
        
        Returns:
            ActionLink -- The link to execute the API action.
        """
        return next(filter(lambda x: x.rel == 'qr-code', self.links), None)

    def to_dict(self) -> dict:
        ret = super().to_dict()        
        if self.items:
            ret['items'] = [ x.to_dict() for x in self.items ]
        if self.refunds:
            ret['refunds'] = [ x.to_dict() for x in self.refunds ]
        if self.payments:
            ret['payments'] = [ x.to_dict() for x in self.payments ]
        if self.primary_recipients:
            ret['primary_recipients'] = [ x.to_dict() for x in self.primary_recipients ]
        return ret

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        primary_recipients, items, payments, refunds, templates = [], [], [], [], []
        detail, invoicer, configuration, amount, due_amount, gratuity  = None, None, None, None, None, None

        if 'primary_recipients' in json_data.keys():
            primary_recipients = [RecipientInfo.serialize_from_json(x, response_type) for x in json_data['primary_recipients']]
        if 'items' in json_data.keys():
            items = [Item.serialize_from_json(x, response_type) for x in json_data['items']]
        if 'payments' in json_data.keys():
            payments = [InvoicePayment.serialize_from_json(x, response_type) for x in json_data['payments']]
        if 'refunds' in json_data.keys():
            refunds = [Refund.serialize_from_json(x, response_type) for x in json_data['refunds']]
        
        if 'detail' in json_data.keys():
            detail = InvoiceDetail.serialize_from_json(json_data['detail'], response_type)
        if 'invoicer' in json_data.keys():
            invoicer = InvoicerInfo.serialize_from_json(json_data['invoicer'], response_type)
        if 'configuration' in json_data.keys():
            configuration = InvoiceConfiguration.serialize_from_json(json_data['configuration'], response_type)
        if 'amount' in json_data.keys():
            amount = Money.serialize_from_json(json_data['amount'], response_type)
        if 'gratuity' in json_data.keys():
            gratuity = Money.serialize_from_json(json_data['gratuity'], response_type)
        if 'due_amount' in json_data.keys():
            due_amount = Money.serialize_from_json(json_data['due_amount'], response_type)
        if 'templates' in json_data.keys():
            templates = [ Template.serialize_from_json(x) for x in json_data['templates'] ]
            
        return cls(
            json_data.get('id'), detail, invoicer, primary_recipients, items, configuration, amount,
            payments, refunds, due_amount, gratuity, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(cls: Type[T], *, detail: InvoiceDetail, invoicer: InvoicerInfo, primary_recipients: List[RecipientInfo], 
        additional_recipients: List[str], items: List[Item], configuration: InvoiceConfiguration, amount: Money, 
        payments: List[InvoicePayment], refunds: List[Refund], due_amount: Money = None, gratuity: Money = None)  -> T:

        add_rep = [{ 'email_address': x} for x in additional_recipients]

        return cls(
            None, detail, invoicer, primary_recipients, items, configuration, amount,
            payments, refunds, due_amount, gratuity, additional_recipients = add_rep
        )

class InvoiceSearchRequest:
    """Search request for invoice queries
    """

    def __init__(
        self, email: str, recipient_first_name: str, recipient_last_name: str,
        recipient_business_name: str, inv_status: List[str], invoice_number: str, reference: str,
        country_code: str, memo: str, total_amount_range: AmountRange, invoice_date_range: DateRange,
        due_date_range: DateRange, payment_date_range: DateRange, creation_date_range: DateRange, archived: bool,
        fields: List[str]
    ):
        self.email = email
        self.recipient_first_name = recipient_first_name
        self.recipient_last_name = recipient_last_name
        self.recipient_business_name = recipient_business_name
        self.inv_status = inv_status
        self.invoice_number = invoice_number
        self.reference = reference
        self.country_code = country_code
        self.memo = memo
        self.total_amount_range = total_amount_range
        self.invoice_date_range     = invoice_date_range
        self.due_date_range = due_date_range
        self.payment_date_range = payment_date_range
        self.creation_date_range = creation_date_range
        self.archived = archived
        self.fields = fields
    
    def to_dict(self) -> dict:
        d = {
            'email' : self.email,
            'recipient_first_name' : self.recipient_first_name,
            'recipient_last_name' : self.recipient_last_name,
            'recipient_business_name' : self.recipient_business_name,
            'inv_status' : self.inv_status,
            'invoice_number' : self.invoice_number,
            'reference' : self.reference,
            'country_code' : self.country_code,
            'memo' : self.memo,
            'total_amount_range' : self.total_amount_range.to_dict(),
            'invoice_date_range' : self.invoice_date_range.to_dict(),
            'due_date_range' : self.due_date_range.to_dict(),
            'payment_date_range' : self.payment_date_range.to_dict(),
            'creation_date_range' : self.creation_date_range.to_dict(),
            'archived' : self.archived,
            'fields' : self.fields
        }

        return { k:v for k,v in d.items() if v != None }
    
    @classmethod
    def create(
        cls, *, email: str = None, recipient_first_name: str = None, recipient_last_name: str = None,
        recipient_business_name: str = None, inv_status: List[str] = None, invoice_number: str = None, reference: str = None,
        country_code: str = None, memo: str = None, total_amount_range: AmountRange = None, invoice_date_range: DateRange = None,
        due_date_range: DateRange = None, payment_date_range: DateRange = None, creation_date_range: DateRange = None, archived: bool = None,
        fields: List[str] = None
    ):
        return cls(
            email, recipient_first_name, recipient_last_name, recipient_business_name, inv_status, invoice_number, reference,
            country_code, memo, total_amount_range, invoice_date_range, due_date_range, payment_date_range, creation_date_range, 
            archived, fields
        )


"""
    Payment resource capture related entities
"""
from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T,
    Money,
    ActionLink,
    PlatformFee,
    ResponseType,    
    PayPalEntity
)

from pypaypal.entities.payments.commons import SellerProtection, ExchangeRate

class CaptureStatusDetailReason(Enum):
    """See https://developer.paypal.com/docs/api/payments/v2/#definition-capture_status_details
    """
    BUYER_COMPLAINT = 1
    CHARGEBACK = 2
    ECHECK = 3
    INTERNATIONAL_WITHDRAWAL = 4
    OTHER = 5
    PENDING_REVIEW = 6
    RECEIVING_PREFERENCE_MANDATES_MANUAL_ACTION = 7
    REFUNDED = 8
    TRANSACTION_APPROVED_AWAITING_FUNDING = 9
    UNILATERAL = 10
    VERIFICATION_REQUIRED = 11

class CaptureStatus(Enum):
    COMPLETED = 1
    DECLINED = 2
    PARTIALLY_REFUNDED = 3
    PENDING = 4
    REFUNDED = 5

class CaptureStatusDetail(PayPalEntity):
    """Capture status details
    """
    def __init__(self, reason: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.reason  = reason

    @property
    def reason_enum(self) -> CaptureStatusDetailReason:
        try:
            return CaptureStatusDetailReason[self.reason] if self.reason else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('reason'), json_response = json_data, response_type = response_type)

    @classmethod
    def create(cls, reason: CaptureStatusDetailReason) -> 'CaptureStatusDetail':
        return cls(reason.name)

class SellerBreakdown(PayPalEntity):
    """Capture seller receivable breakdown    
    """
    def __init__(
        self, gross_amt: Money, paypal_fee: Money, net_amt: Money, 
        receivable_amt: Money, exchange_rate: ExchangeRate, 
        platform_fees: List[PlatformFee], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.gross_amount = gross_amt
        self.paypal_fee = paypal_fee
        self.net_amount = net_amt
        self.receivable_amount = receivable_amt
        self.exchange_rate  = exchange_rate
        self.platform_fees = platform_fees

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        platform_fees = []
        gross_amount, paypal_fee, net_amount, receivable_amount, exchange_rate = None, None, None, None, None

        if 'gross_amount' in json_data.keys():
            gross_amount = Money.serialize_from_json(json_data['gross_amount'], response_type)
        if 'paypal_fee' in json_data.keys():
            paypal_fee = Money.serialize_from_json(json_data['paypal_fee'], response_type)
        if 'net_amount' in json_data.keys():
            net_amount = Money.serialize_from_json(json_data['net_amount'], response_type)
        if 'receivable_amount' in json_data.keys():
            receivable_amount = Money.serialize_from_json(json_data['receivable_amount'], response_type)
        
        if 'platform_fees' in json_data.keys():
            platform_fees = [PlatformFee.serialize_from_json(x, response_type) for x in json_data['platform_fees']]
        
        return cls(
            gross_amount, paypal_fee, net_amount, receivable_amount, 
            exchange_rate, platform_fees, json_response = json_data, 
            response_type = response_type
        )

class Capture(PayPalEntity):
    """Payment capture obj representation.
    """
    def __init__(
        self, cap_id: str, status: str, final_capture: bool, 
        disbursement_mode: str, invoice_id: str, custom_id: str,
        amount: Money, seller_protection: SellerProtection,
        status_details: CaptureStatusDetail , seller_breakdown: SellerBreakdown, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = cap_id
        self.status = status
        self.final_capture  = final_capture
        self.disbursement_mode = disbursement_mode
        self.invoice_id = invoice_id
        self.custom_id = custom_id
        self.amount = amount
        self.seller_protection = seller_protection
        self.status_details = status_details
        self.seller_receivable_breakdown = seller_breakdown
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]
    
    @property
    def status_enum(self) -> CaptureStatus:
        try:
            return CaptureStatus[self.status] if self.status else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def Status_enum(self) -> CaptureStatusDetailReason:
        try:
            return CaptureStatusDetailReason[self.status] if self.status else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)
    
    @property
    def up_link(self) -> ActionLink:
        """Retrieves a link to read this entity authorization details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'up', self.links), None)
    
    @property
    def refund_link(self) -> ActionLink:
        """Retrieves a link to refund this capture.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'refund', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount, seller_protection, status_detail, seller_receivable_breakdown = None, None, None, None
        
        if 'amount' in json_data.keys():
            amount = cls.serialize_from_json(json_data['amount'], response_type)
        if 'seller_protection' in json_data.keys():
            seller_protection = cls.serialize_from_json(json_data['seller_protection'], response_type)
        if 'status_details' in json_data.keys():
            status_detail = cls.serialize_from_json(json_data['status_details'], response_type)
        if 'seller_receivable_breakdown' in json_data.keys():
            seller_receivable_breakdown = cls.serialize_from_json(json_data['seller_receivable_breakdown'], response_type)

        return cls(
            json_data.get('id'), json_data.get('status'), json_data.get('final_capture'), 
            json_data.get('disbursement_mode'), json_data.get('invoice_id'), json_data.get('custom_id'), 
            amount, seller_protection, status_detail, seller_receivable_breakdown, json_response = json_data,
            response_type = response_type
        )


"""
    Module with all product related entities 
"""

from enum import Enum
from typing import Type
from datetime import datetime

import dateutil.parser
from pypaypal.entities.base import T, PayPalEntity, ResponseType, ActionLink, PatchUpdateRequest

class ProductType(Enum):
    PHYSICAL = 1
    DIGITAL = 2
    SERVICE = 3

class UpdateAction(Enum):
    ADD = 1
    REMOVE = 2
    REPLACE = 3

class UpdateField(Enum):
    DESCRIPTION = 1
    CATEGORY = 2
    IMAGE_URL = 3
    HOME_URL = 4

class ProductUpdateRequest(PatchUpdateRequest):
    """Request object for updating a product
    """

    def __init__(self, field: UpdateField, action: UpdateAction, value: str):
        super().__init__('/{}'.format(field.name.lower()), value, action.name.lower())
            
class Product(PayPalEntity):
    """
        Object representation of a paypal product
    """
    def __init__(self, name: str, product_type: ProductType, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.product_type = product_type
        self.id = self._json_response.get('id', kwargs.get('id')) 
        self.category = self._json_response.get('category', kwargs.get('category'))
        self.home_url = self._json_response.get('home_url', kwargs.get('home_url'))
        self.image_url = self._json_response.get('image_url', kwargs.get('image_url'))
        self.description = self._json_response.get('description', kwargs.get('description'))
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def update_link(self) -> ActionLink:
        """Retrieves the update link for this entity
        
        Returns:
            ActionLink -- The link for requesting an update in the API.
        """
        return next(filter(lambda x: x.rel == 'edit', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data['name'], ProductType[json_data['type']], json_response= json_data, response_type = response_type
        )


"""
    Module with all Payment Expirience Web Profiles related entities.
"""

from enum import Enum
from typing import Type, List

from pypaypal.entities.base import (
    T, 
    PayPalEntity, 
    ResponseType
)

class LandingPageType(Enum):
    BILLING = 1
    LOGIN = 2

class UserAction(Enum):
    CONTINUE = 1
    COMMIT = 2

class ShippingOpt(Enum):
    DISPLAY = 0
    REDACT = 1
    GET_FROM_PROFILE = 2

class AddressOverrideOpt(Enum):
    DISPLAY_FROM_FILE = 0
    DISPLAY_FROM_CALL = 1

class FlowConfig(PayPalEntity):
    """Webexp Flow configuration object representation.
    """

    def __init__(
        self, landing_page_type: str, bank_txn_pend_url: str, 
        user_action: str, ret_uri_http_meth: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.landing_page_type = landing_page_type
        self.user_action = user_action
        self.bank_txn_pending_url = bank_txn_pend_url
        self.return_uri_http_method = ret_uri_http_meth
    
    @property
    def user_action_enum(self) -> UserAction:
        try:
            return UserAction[self.user_action] if self.user_action else None
        except:
            return None

    @property
    def landing_page_type_enum(self) -> LandingPageType:
        try:
            return LandingPageType[self.landing_page_type] if self.landing_page_type else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        return cls(
            json_data.get('landing_page_type'), json_data.get('bank_txn_pending_url'), 
            json_data.get('user_action'), json_data.get('return_uri_http_method'), 
            json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def create(
        cls, *, landing_page_type: LandingPageType = LandingPageType.BILLING, 
        bank_txn_pending_url: str = None, user_action: UserAction = UserAction.CONTINUE, 
        return_uri_http_method: str = None) -> 'FlowConfig':
        ua = 'continue' if user_action == UserAction.CONTINUE else 'commit'
        lpt = 'Login' if landing_page_type == LandingPageType.LOGIN else 'Billing'
        return cls(lpt, bank_txn_pending_url, ua, return_uri_http_method)

class InputField(PayPalEntity):
    """Input field obj representation
    """

    def __init__(self, no_shipping: int, address_override: int, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.no_shipping = no_shipping
        self.address_override = address_override
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:        
        return cls(
            json_data.get('no_shipping'), json_data.get('address_override'),
            json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def create(cls, no_shipping: ShippingOpt, address_override: AddressOverrideOpt) -> 'InputField':
        return cls(no_shipping.value, address_override.value)

class Presentation(PayPalEntity):
    """WebExp Page Presentation obj representation
    """
    def __init__(self, brand_name: str, logo_image: str, locale_code: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.brand_name = brand_name
        self.logo_image = logo_image
        self.locale_code = locale_code

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('brand_name'), json_data.get('logo_image'), 
            json_data.get('locale_code'), json_response= json_data, 
            response_type = response_type
        )

    @classmethod
    def create(cls, brand_name: str, logo_url: str, locale_code: str) -> 'Presentation':
        return cls(brand_name, logo_url, locale_code)

class WebExpProfile(PayPalEntity):
    """Web Expirience Profile object representation
    """
    def __init__(
        self, name: str, temporary: bool, flow_config: FlowConfig, 
        input_fields: InputField, presentation: Presentation, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.temporary = temporary
        self.flow_config  = flow_config
        self.input_fields = input_fields
        self.presentation = presentation
        self.id = self._json_response.get('id', kwargs.get('id'))
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        flow_config, input_fields, presentation = None, None, None

        if 'flow_config' in json_data.keys():
            flow_config = FlowConfig.serialize_from_json(json_data['flow_config'], response_type)
        if 'input_fields' in json_data.keys():
            input_fields = InputField.serialize_from_json(json_data['input_fields'], response_type)
        if 'presentation' in json_data.keys():
            presentation = Presentation.serialize_from_json(json_data['presentation'], response_type)
        
        return cls(json_data.get('name'), json_data.get('temporary'), flow_config, input_fields, presentation)
    
    @classmethod
    def create(
        cls, name: str, temporary: bool, flow_config: FlowConfig = None,
        input_fields: InputField = None, presentation: Presentation = None) -> 'WebExpProfile':
        return cls(name, temporary, flow_config, input_fields, presentation)


"""
Copyright (C) 2020 Ivan Cuello

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author:
    Ivan Cuello <ivcuello@gmail.com>
"""

import os
import sys

import pathlib
from setuptools import setup, find_packages

# Current folder and README.md
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

CURRENT_PYTHON_VERSION = sys.version_info[:2]
REQUIRED_PYTHON_VERSION = (3,6)

# Validating python version before the setup run
if REQUIRED_PYTHON_VERSION > CURRENT_PYTHON_VERSION:
    sys.stderr.write("""
    Unsupported python version found.

    Current Python: {}.
    Required Python: {}.
    
    """.format(CURRENT_PYTHON_VERSION, REQUIRED_PYTHON_VERSION))
    sys.exit(64)

setup(
    name='pypaypal',
    version='1.0.0',
    python_requires='>=3.6',
    author='ivcuello',
    author_email='ivcuello@gmail.com',
    url='https://github.com/ivcuello/pypaypal',
    description='Paypal API integration supporting some v1 & most of the current v2 rest APIs calls',
    long_description=README,
    long_description_content_type="text/markdown",    
    packages = find_packages(exclude=['docs', 'tests']),
    install_requires = [ 
        'python-dateutil',
        'requests'
    ],
    license="Apache License 2.0",
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ]
)



"""
    Paypal Webhook resource client for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/webhooks/v1/
"""

import json

from datetime import datetime
from typing import TypeVar, Set, List

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL, 
    LEGACY_SANDBOX_API_BASE_URL 
)

from pypaypal.entities.base import ( 
    DateRange,
    PaypalPage,
    ResponseType,
    PaypalApiResponse,
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.webhooks import (
    Webhook,
    EventType,
    AnchorType,
    WebhookEvent,
    WebhookSignature,
    SignatureVerificationStatus
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'notifications', 'webhooks')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'notifications', 'webhooks')

# Webhook client
W = TypeVar('W', bound = 'WebHookClient')

# Webhook signature verification client
V = TypeVar('S', bound = 'VerifySignatureClient')

# Webhook event type client
T = TypeVar('T', bound = 'EventTypeClient')

# Webhook events client
E = TypeVar('E', bound = 'EventClient')

# Webhook evebt simulation client
S = TypeVar('S', bound = 'EventSimulationClient')

class ApiCallError(Exception):
    """Exception for methods that don't return a PaypalEntity
    """
    def __init__(self, error: PaypalApiResponse):
        super().__init__(json.dumps(ApiCallError._parse_error(error)))
        self.error_info = error
    
    @staticmethod
    def _parse_error(error: PaypalApiResponse) -> dict:
        return error.error_detail.as_json_dict() if error.error_detail else {'message': 'unknown error'}

class WebHookClient(ClientBase):
    """WebHook resource client
    """
    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)

    def _process_simple_response(self, api_response) -> PaypalApiResponse[Webhook]:
        """Process a simple webhook response
        
        Arguments:
            api_response {[type]} -- The http response obj from the request lib
        
        Returns:
            PaypalApiResponse[Webhook] -- Paypal response status with webhook object.
        """
        if api_response.status_code//100 != 2:
            return PaypalApiResponse.error(api_response)

        return PaypalApiResponse.success(api_response, Webhook.serialize_from_json(api_response.json()))

    
    def create_webhook(self, url: str, event_types: List[EventType], subscribe_all: bool = False) -> PaypalApiResponse[Webhook]:
        """Calls the paypal API to create a webhook
        
        Arguments:
            url {str} --  Internet accessible URL configured to listen incoming POST notification messages with event info.
            event_types {List[EventType]} -- An array of events to which to subscribe your webhook.
        
        Returns:
            PaypalApiResponse[Webhook] -- Paypal response status with webhook object.
        """
        body = json.dumps({
            'url': url, 
            'event_types': [x.to_dict() for x in event_types] if not subscribe_all else '*'
        })
        
        return self._process_simple_response(self._session.post(self._base_url, body))


    def list_webhooks(self, anchor_type: AnchorType = AnchorType.APPLICATION) -> PaypalApiBulkResponse[Webhook]:
        """Calls the paypal api to list an app webhooks
        
        Arguments:
            anchor_type {AnchorType} -- AnchorType filter
        
        Returns:
            PaypalApiBulkResponse[Webhook] -- Paypal response status with webhook objects.
        """
        api_response = self._session.get(self._base_url, params = {'anchor_type': anchor_type.name})

        if api_response.status_code//100 != 2:
            return PaypalApiBulkResponse.error(api_response)

        return PaypalApiBulkResponse.success(
            api_response, 
            [ Webhook.serialize_from_json(x) for x in api_response.json().get('webhooks', []) ]
        )
    
    def delete_webhook(self, webhook_id: str) -> PaypalApiResponse:
        """Calls the paypal API to delete a webhook by id.
        
        Arguments:
            webhook_id {str} -- Id of the Webhook to be deleted. 
        
        Returns:
            PaypalApiResponse -- Paypal response status
        """
        api_response = self._session.delete(parse_url(self._base_url, webhook_id))
        return PaypalApiResponse.error(api_response) if api_response.status_code//100 != 2 else PaypalApiResponse.success(api_response)

    def update_webhook(self, webhook_id: str, patch_request: List[PatchUpdateRequest]) -> PaypalApiResponse[Webhook]:
        """Calls the paypal API to replace webhook fields with new values
        
        Arguments:
            webhook_id {str} -- Id of the webhook to be updated
            patch_request {List[PatchUpdateRequest]} -- patch request list with updates
        
        Returns:
            PaypalApiResponse[Webhook] -- Paypal response status with webhook object. 
        """
        body = json.dumps({'patch_request': [ x.to_dict() for x in patch_request ]})
        return self._process_simple_response(self._session.patch(parse_url(self._base_url, webhook_id), body))

    def show_webhook_details(self, webhook_id: str) -> PaypalApiResponse[Webhook]:
        """Calls the paypal API to show a webhook details
        
        Arguments:
            webhook_id {str} -- Webhook id
        
        Returns:
            PaypalApiResponse[Webhook] -- Response status with webhook object.
        """
        return self._process_simple_response(self._session.get(parse_url(self._base_url, webhook_id)))

    def list_event_subscriptions_for_webhook(self, webhook_id: str) -> PaypalApiBulkResponse[EventType]:
        """Calls the paypal API to list the event subscription for a Webhook
        
        Arguments:
            webhook_id {str} -- Webhook id
        
        Returns:
            PaypalApiBulkResponse[EventType] -- Paypal response status with event type list. 
        """
        api_response = self._session.get(parse_url(self._base_url, webhook_id, 'event-types'))

        if api_response.status_code//100 != 2:
            return PaypalApiBulkResponse.error(api_response)

        return PaypalApiBulkResponse.success(
            api_response, 
            [ EventType.serialize_from_json(x) for x in api_response.json().get('event_types', []) ]
        )

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> W:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'webhooks'), session)

class VerifySignatureClient(ClientBase):
    """Verify WebHook Signatire resource client
    """
    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)

    def verify_webhook_signature(self, signature: WebhookSignature) -> SignatureVerificationStatus:
        """Calls the paypal API to verify a webhook signature (it usually comes in the headers).
        
        Arguments:
            signature {WebhookSignature} -- The signature
        
        Returns:
            SignatureVerificationStatus -- Verification status
        
        Raises:
            ApiCallError -- If there's an error calling the API.
        """        
        api_response = self._session.post(self._base_url, json.dumps(signature.to_dict()))

        if api_response.status_code//100 != 2:
            raise ApiCallError(PaypalApiResponse.error(api_response))

        return SignatureVerificationStatus[api_response.json().get('verification_status')]

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> V:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'verify-webhook-signature'), session)
    
class EventTypeClient(ClientBase):
    """Verify WebHook Signatire resource client
    """
    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)

    def list_available_events(self) -> PaypalApiBulkResponse[EventType]:
        """Calls the paypal API to get a list of available events to which any webhook can subscribe.
        
        Returns:
            PaypalApiBulkResponse[EventType] -- Paypal response status with the event type list.
        """
        api_response = self._session.get(self._base_url)

        if api_response.status_code//100 != 2:
            return PaypalApiBulkResponse.error(api_response)

        return PaypalApiBulkResponse.success(
            api_response, 
            [ EventType.serialize_from_json(x) for x in api_response.json().get('event_types', []) ]
        )

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'webhooks-event-types'), session)
    
class EventClient(ClientBase):
    """Verify WebHook Signatire resource client
    """
    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)

    def list_event_notifications(
            self, page_size: int = 10, transaction_id: str = None, 
            event_type: str = None, date: DateRange = None
        ) -> PaypalPage[WebhookEvent]:
        params = { 'page_size': 10 }

        if transaction_id != None: 
            params['transaction_id'] = transaction_id
        if event_type != None: 
            params['event_type'] = event_type
        if date != None and date.start_time != None: 
            params['start_time'] = datetime.strftime(date.start, '%Y-%m-%dT%H:%M:%S') 
        if date != None and date.end_time != None:
            params['end_time'] = datetime.strftime(date.end, '%Y-%m-%dT%H:%M:%S')
        
        api_response = self._session.get(self._base_url, params = params)

        if api_response.status_code // 100 != 2:
            return PaypalPage.error(api_response)
        
        return PaypalPage.full_parse_success(api_response, WebhookEvent, 'events')

    def show_event_notification_details(self, event_id: str) -> PaypalApiResponse[WebhookEvent]:
        """Calls the API to get the details for a webhook event notification.
        
        Arguments:
            event_id {str} -- The webhook notification event id
        
        Returns:
            PaypalApiResponse[WebhookEvent] -- Paypal response status with event obj.
        """
        api_response = self._session.get(parse_url(self._base_url, event_id))

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, WebhookEvent.serialize_from_json(api_response.json()))

    def resend_event_notification(self, event_id: str, webhooks_ids: List[str] = []) -> PaypalApiResponse[WebhookEvent]:
        """Calls the paypal API to resend notification events
        
        Arguments:
            event_id {str} -- The webhook event id
        
        Keyword Arguments:
            webhooks_ids {List[str]} -- webhooks notification to resend (all pending if empty) (default: {[]})
        
        Returns:
            PaypalApiResponse[WebhookEvent] -- A response with the webhooks events.
        """
        body = { 'webhook_ids': webhooks_ids or [] }
        api_response = self._session.post(parse_url(self._base_url, event_id), json.dumps(body))

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, WebhookEvent.serialize_from_json(api_response.json()))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> E:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'webhooks-events'), session)

class SimulateEventClient(ClientBase):
    """Verify WebHook Signatire resource client
    """
    def __init__(self, base_url: str, session: PayPalSession):
        super().__init__(base_url, session)

    def simulate_webhook_event(self, webhook_id: str, url: str, event_type: str, resource_version: str) -> PaypalApiResponse[WebhookEvent]:
        """Calls the paypal API to send a webhook simulation
        
        Arguments:
            webhook_id {str} -- Webhook id, if ommited url is required
            url {str} -- endpoint url, if ommited id is required 
            event_type {str} -- The event name, one per request
            resource_version {str} -- event type version
        
        Returns:
            PaypalApiResponse[WebhookEvent] -- A response with the webhooks events.
        """
        body = { 'event_type': event_type }
        
        if body['url'] != None: 
            body['url'] = url
        if body['webhook_id'] != None: 
            body['webhooks_id'] = webhook_id
        if body['resource_version'] != None: 
            body['resource_version'] = resource_version

        api_response = self._session.post(self._base_url, json.dumps(body))

        if api_response.status_code // 100 != 2:
            return PaypalApiResponse.error(api_response)
        
        return PaypalApiResponse.success(api_response, WebhookEvent.serialize_from_json(api_response.json()))


    @classmethod
    def for_session(cls: T, session: PayPalSession) -> E:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'simulate-event'), session)



"""
    Module with all Order related entities.
"""

from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import ( 
    T, 
    Money,
    Payee,
    PaypalName,
    ActionLink,
    ResponseType,
    PayPalEntity,
    PaymentMethod,
    PaypalPhoneDetail,
    AppCtxLandingPage,
    ApplicationContext,
    PaymentInstruction,
    PaypalPortableAddress,
    AppCtxShippingPreference
)

from pypaypal.entities.payments.captures import Capture
from pypaypal.entities.payments.authorizations import Authorization

class PhoneType(Enum):
    FAX = 1
    HOME = 2
    MOBILE = 3
    OTHER = 4
    PAGER = 5

class OrderIntent(Enum):
    CAPTURE = 1
    AUTHORIZE = 2

class TaxIdType(Enum):
    BR_CPF = 1 # Individual tax id type
    BR_CNPJ = 2 # Business tax id type

class ItemCategory(Enum):
    DIGITAL_GOODS = 1
    PHYSICAL_GOODS = 2

class OrderStatus(Enum):
    CREATED = 1
    SAVED = 2
    APPROVED = 3
    VOIDED = 4
    COMPLETED = 5

class OrderUserAction(Enum):
    CONTINUE = 1
    PAY_NOW = 2

class CardType(Enum): 
    VISA = 1 # Visa card.
    MASTERCARD = 2  # MasterCard card.
    DISCOVER = 3 # Discover card.
    AMEX = 4 # American Express card.
    SOLO = 5 # Solo debit card.
    JCB = 6 # Japan Credit Bureau card.
    STAR = 7 # Military Star card.
    DELTA = 8 # Delta Airlines card.
    SWITCH = 9 # Switch credit card.
    MAESTRO = 10 # Maestro credit card.
    CB_NATIONALE = 11 # Carte Bancaire (CB) credit card.
    CONFIGOGA = 12 # Configoga credit card.
    CONFIDIS = 13 # Confidis credit card.
    ELECTRON = 14 # Visa Electron credit card.
    CETELEM = 15 # Cetelem credit card.
    CHINA_UNION_PAY = 16 # China union pay credit card.

class TokenType(Enum):
    BILLING_AGREEMENT = 1 # Approved recurring payment for goods or services.

class TypedPhone(PayPalEntity):
    """Paypal Phone with type object representation.
    """
    def __init__(self, phone_type: str, phone_number: PaypalPhoneDetail, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.phone_type = phone_type
        self.phone_number = phone_number
    
    def to_dict(self) -> dict:
        d = super().to_dict()
        d['phone_number'] = { 'national_number': self.phone_number.national_number }
        return d

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        phone_number = None

        if 'phone_number' in json_data.keys():
            phone_number = PaypalPhoneDetail.serialize_from_json(json_data['phone_number'], response_type)
        
        return cls(
            json_data.get('phone_type'), phone_number, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(cls, phone_type: PhoneType, phone_number: PaypalPhoneDetail) -> 'TypedPhone':
        return cls(phone_type, phone_number)

class TaxInfo(PayPalEntity):
    """Payer tax info object representation
    """
    def __init__(self, tax_id: str, tax_id_type: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.tax_id = tax_id
        self.tax_id_type = tax_id_type
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(
            json_data.get('tax_id'), json_data.get('tax_id_type'), json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(cls, tax_id: str, tax_id_type: TaxIdType) -> 'TaxInfo':
        return cls(tax_id, tax_id_type.name)

class Payer(PayPalEntity):
    """Paypal Payer object representation
    """
    def __init__(
        self, payer_id: str, name: PaypalName, email_addr: str, 
        phone: TypedPhone, birth_date: str, tax_info: TaxInfo, 
        address: PaypalPortableAddress, **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))    
        self.payer_id = payer_id
        self.name = name
        self.email_address = email_addr 
        self.phone = phone
        self._birth_date = birth_date
        self.tax_info = tax_info 
        self.address = address

    def to_dict(self) -> dict:
        d = super().to_dict()
        if '_birth_date' in d.keys():
            d['birth_date'] = d.pop('_birth_date')
        return d

    @property
    def birth_date(self) -> datetime:
        try:
            return datetime.strptime(self._birth_date, '%Y-%m-%d') if self._birth_date != None else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        name, phone, tax_info, address = None, None, None, None

        if 'name' in json_data.keys():
            name = PaypalName.serialize_from_json(json_data['name'], response_type)
        if 'phone' in json_data.keys():
            phone = TypedPhone.serialize_from_json(json_data['phone'], response_type)
        if 'tax_info' in json_data.keys():
            tax_info = TaxInfo.serialize_from_json(json_data['tax_info'], response_type)
        if 'address' in json_data.keys():
            address = PaypalPortableAddress.serialize_from_json(json_data['address'], response_type)

        return cls(
            json_data.get('payer_id'), name, json_data.get('email_address'), phone, 
            json_data.get('birth_date'), tax_info, address, json_response= json_data, 
            response_type = response_type
        )

    @classmethod
    def create(
        cls, payer_id: str, name: PaypalName = None, email_addr: str = None,
        phone: TypedPhone = None, birth_date: str = None, tax_info: TaxInfo  = None, 
        address: PaypalPortableAddress = None
    ) -> 'Payer':
        return cls(payer_id, name, email_addr, phone, birth_date, tax_info, address)

class AmountBreakdown(PayPalEntity):
    """Amount breakdown detail
    """
    def __init__(
        self, item_total: Money, shipping: Money, handling: Money, 
        tax_total: Money, insurance: Money, shipping_discount: Money,
        discount: Money, **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))    
        self.item_total = item_total
        self.shipping = shipping
        self.handling  = handling 
        self.tax_total = tax_total
        self.insurance = insurance
        self.shipping_discount = shipping_discount
        self.discount = discount
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        breakdown = { 
            'item_total': None, 'shipping': None, 'handling': None, 
            'tax_total': None, 'insurance': None, 'shipping_discount': None, 
            'discount': None
        }

        breakdown = { k : Money.serialize_from_json(json_data[k], response_type) if k in json_data.keys() else None for k in breakdown }

        return cls(**breakdown, json_response = json_data, response_type = response_type)
    
    @classmethod
    def create(cls, *, item_total: Money = None, shipping: Money = None, handling: Money = None, 
        tax_total: Money = None, insurance: Money = None, shipping_discount: Money = None,
        discount: Money = None):
        return cls(item_total, shipping, handling, tax_total, insurance, shipping_discount, discount)

class AmountWithBreakdown(PayPalEntity):
    """Amount with breakdown object representations for order entities
    """
    def __init__(self, currency_code: str, value: str, breakdown: AmountBreakdown, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.currency_code = currency_code
        self.value = value
        self.breakdown = breakdown
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        breakdown = AmountBreakdown.serialize_from_json(json_data['breakdown'], response_type) if 'breakdown' in json_data.keys() else None
        return cls(
            json_data.get('currency_code'), json_data.get('value'), breakdown,
            json_response = json_data, response_type = response_type
        )
    
    @classmethod
    def create(cls, *, currency_code: str, value: str, breakdown: AmountBreakdown = None):
        return cls(currency_code, value, breakdown)

class Item(PayPalEntity):
    """Order item obj representation. Yes... it's different from the Item class in the base module
    """
    def __init__(
        self, name: str, unit_amount: Money, tax: Money, quantity: str, 
        category: str, description: str = None, sku: str = None, **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.unit_amount = unit_amount
        self.tax = tax
        self.quantity = quantity
        self.category = category
        self.description = description
        self.sku = sku

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        unit_amount, tax = None, None

        if 'unit_amount' in json_data.keys():
            unit_amount = Money.serialize_from_json(json_data['unit_amount'], response_type)
            
        if 'tax' in json_data.keys():
            tax = Money.serialize_from_json(json_data['tax'], response_type)
        
        return cls(
            json_data.get('name'), unit_amount, tax, json_data.get('quantity'),
            json_data.get('category'), json_data.get('description'), json_data.get('sku'),
            json_response= json_data, response_type = response_type
        )
    
    @classmethod
    def create(
        cls, *, name: str, unit_amount: Money, quantity: int, 
        category: ItemCategory, tax: Money = None, description: str = None, 
        sku: str = None
    ):
        return cls(name, unit_amount, tax, quantity, category.name, description, sku)

class Name(PayPalEntity):
    """Order name obj representation.
    """
    def __init__(self, full_name: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.full_name = full_name

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('full_name'), json_response= json_data, response_type = response_type)

class ShippingDetail(PayPalEntity):
    """Shipping detail obj representation
    """

    def __init__(self, name: Name, address: PaypalPortableAddress, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.address = address
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        name, address = None, None

        if 'name' in json_data.keys():
            name = Name.serialize_from_json(json_data['name'], response_type)
        if 'address' in json_data.keys():
            address = PaypalPortableAddress.serialize_from_json(json_data['address'], response_type)
        
        return cls(name, address, json_response= json_data, response_type = response_type)
    
    @classmethod
    def create(cls, full_name: str, address: PaypalPortableAddress) -> 'ShippingDetail':
        return cls(Name(full_name), address)

class OrderPayment(PayPalEntity):
    """Payment attr for orders
    """
    def __init__(self, authorizations: List[Authorization], captures: List[Capture], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.captures = captures
        self.authorizations = authorizations
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        captures, authorizations = None, None

        if 'captures' in json_data.keys():
            captures = [Capture.serialize_from_json(x, response_type) for x in json_data['captures']]
        if 'authorizations' in json_data.keys():
            authorizations = [Authorization.serialize_from_json(x, response_type) for x in json_data['authorizations']]

        return cls(authorizations, captures, json_response= json_data, response_type = response_type)

class PurchaseUnitRequest(PayPalEntity):
    """Purchase unit object representation.
    """
    def __init__(
        self, reference_id: str, amount: AmountWithBreakdown, payee: Payee, 
        payment_instruction: PaymentInstruction, description: str, custom_id: str,
        invoice_id: str, pur_id: str, soft_descriptor: str, items: List[Item], 
        shipping: ShippingDetail, payments: OrderPayment = None, **kwargs
    ):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.reference_id = reference_id
        self.amount = amount
        self.payee  = payee
        self.payments = payments
        self.payment_instruction = payment_instruction
        self.description = description
        self.custom_id = custom_id
        self.invoice_id = invoice_id
        self.id = pur_id
        self.soft_descriptor = soft_descriptor
        self.items  = items
        self.shipping = shipping

    @property
    def captures(self) -> List[Capture]:
        return self.payments.captures if self.payments and self.payments.captures else []

    @property
    def authorizations(self) -> List[Authorization]:
        return self.payments.authorizations if self.payments and self.payments.authorizations else []

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        payee, amount, payment_instruction, shipping, items, payments = None, None, None, None, [], None
        
        if 'payee' in json_data.keys():
            payee = Payee.serialize_from_json(json_data['payee'], response_type)
        if 'amount' in json_data.keys():
            amount = AmountWithBreakdown.serialize_from_json(json_data['amount'], response_type)
        if 'payment_instruction' in json_data.keys():
            payment_instruction = PaymentInstruction.serialize_from_json(json_data['payment_instruction'], response_type)
        if 'shipping' in json_data.keys():
            shipping = ShippingDetail.serialize_from_json(json_data['shipping'], response_type)
        if 'payments' in json_data.keys():
            payments = OrderPayment.serialize_from_json(json_data['payments'], response_type)

        return cls(
            json_data.get('reference_id'), amount, payee, payment_instruction, json_data.get('description'), 
            json_data.get('custom_id'), json_data.get('invoice_id'), json_data.get('id'), json_data.get('soft_descriptor'), 
            items, shipping, payments, json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(
        cls, amount: AmountWithBreakdown, *, reference_id: str = None, payee: Payee = None,
        payment_instruction: PaymentInstruction = None, description: str = None, custom_id: str = None,
        invoice_id: str = None, pur_id: str = None, soft_descriptor: str = None, items: List[Item] = [],
        shipping: ShippingDetail = None
    ):
        return cls(
            reference_id, amount, payee, payment_instruction, description, custom_id, 
            invoice_id, pur_id, soft_descriptor, items or [], shipping
        )

class OrderApplicationContext(ApplicationContext):
    """Paypal Order app context object representation
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    
    @property
    def order_user_action(self) -> OrderUserAction:
        try:
            return OrderUserAction[self.user_action] if self.user_action else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return super().serialize_from_json(json_data, response_type)
            
    @classmethod
    def create(cls, brand_name: str = None, locale: str = None, return_url: str = None, cancel_url: str = None, 
        payment_method: PaymentMethod = None, landing_page: AppCtxLandingPage = AppCtxLandingPage.NO_PREFERENCE,
        shipping_preference: AppCtxShippingPreference = AppCtxShippingPreference.GET_FROM_FILE, 
        user_action: OrderUserAction = OrderUserAction.CONTINUE) -> 'ApplicationContext':
        return super().create(
            brand_name, locale, landing_page.name, shipping_preference.name,
            user_action.name, payment_method, return_url, cancel_url
        )

class Order(PayPalEntity):
    """Paypal Order object representation
    """

    def __init__(self, intent: str, payer: Payer, purchase_units: List[PurchaseUnitRequest], application_context: OrderApplicationContext, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.intent = intent
        self.payer = payer
        self.purchase_units = purchase_units
        self.application_context = application_context
        self.status = self._json_response.get('status')
        self.id = self._json_response.get('id', kwargs.get('id'))
        self._create_time = self._json_response.get('create_time')
        self._update_time = self._json_response.get('update_time')
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def status_enum(self) -> OrderStatus:
        """Status of the order as an enum constant
        
        Returns:
            OrderStatus -- An enumerated constant representing the order status or None
        """
        try:
            return OrderStatus[self.status] if self.status else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return datetime.strptime(self._create_time, '%Y-%m-%d') if self._create_time != None else None
        except:
            return None

    @property
    def update_time(self) -> datetime:
        try:
            return datetime.strptime(self._update_time, '%Y-%m-%d') if self._update_time != None else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def approve_link(self) -> ActionLink:
        """Retrieves a link to make an approval on this order.
        
        Returns:
            ActionLink -- The link for the API action.
        """
        return next(filter(lambda x: x.rel == 'approve', self.links), None)

    @property
    def capture_link(self) -> ActionLink:
        """Retrieves a link to capture a payment for this order.
        
        Returns:
            ActionLink -- The link for the API action.
        """
        return next(filter(lambda x: x.rel == 'capture', self.links), None)

    def to_dict(self) -> dict:
        d = super().to_dict()
        if '_birth_date' in d.keys():
            d['birth_date'] = d.pop('_birth_date')
        return d

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        payer, application_context, purchase_units = None, None, []

        if 'payer' in json_data.keys():
            payer = Payer.serialize_from_json(json_data['payer'], response_type)
        
        if 'application_context' in json_data.keys():
            application_context = OrderApplicationContext.serialize_from_json(json_data['application_context'], response_type)
        
        if 'purchase_units' in json_data.keys():
            purchase_units = [PurchaseUnitRequest.serialize_from_json(x, response_type) for x in json_data['purchase_units']]

        return cls(
            json_data.get('intent'), payer, purchase_units, application_context,
            json_response= json_data, response_type = response_type
        )

    @classmethod
    def create(cls, *, intent: OrderIntent, purchase_units: List[PurchaseUnitRequest], payer: Payer = None, application_context: OrderApplicationContext = None) -> 'Order':
        return cls(intent.name, payer, purchase_units, application_context)

class Card(PayPalEntity):
    """PaymentSource.card object representation.
    """
    def __init__(
        self, card_id: str, name: str, number: str, expiry: str, 
        security_code: str, last_digits: str, card_type: str, 
        billing_address: PaypalPortableAddress, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.name = name
        self.id = card_id
        self.number = number
        self.expiry = expiry
        self.security_code = security_code
        self.last_digits = last_digits
        self.card_type = card_type
        self.billing_address = billing_address

    @property
    def card_type_enum(self) -> OrderStatus:
        """Status of the order as an enum constant
        
        Returns:
            OrderStatus -- An enumerated constant representing the order status or None
        """
        try:
            return CardType[self.card_type] if self.card_type else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        billing_address = None

        if 'billing_address' in json_data.keys():
            billing_address = PaypalPortableAddress.serialize_from_json(json_data['billing_address'], response_type)
        
        return cls(
            json_data.get('id'), json_data.get('name'), json_data.get('number'),
            json_data.get('expiry'), json_data.get('security_code'), json_data.get('last_digits'),
            json_data.get('card_type'), billing_address, json_response= json_data,
            response_type = response_type
        )

    @classmethod
    def create(
        cls, number: str, expiry: str, *, name: str = None, 
        security_code: str = None, billing_address: PaypalPortableAddress = None) -> 'Card':
        # CardType will remain as "str" to keep flexibilty in case there's another supported network.
        return cls(None, name, number, expiry, security_code, None, None, billing_address)

class Token(PayPalEntity):
    """PaymentSource.token object representation.
    """
    def __init__(self, token_id: str, token_type: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.id = token_id
        self.type= token_type

    @property
    def token_type_enum(self) -> OrderStatus:
        """Status of the order as an enum constant
        
        Returns:
            OrderStatus -- An enumerated constant representing the order status or None
        """
        try:
            return TokenType[self.type] if self.type else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('id'), json_data.get('type'), json_response= json_data, response_type = response_type)
    
    @classmethod
    def create(cls, token_id: str, token_type: str = TokenType.BILLING_AGREEMENT.name) -> 'Token':
        return cls(token_id, token_type)

class PaymentSource(PayPalEntity):
    """Payment source object representation.
    """
    def __init__(self, card: Card, token: Token, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.card = card
        self.token = token
    
    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        card, token = None, None

        if 'card' in json_data.keys():
            card = Card.serialize_from_json(json_data['card'], response_type)
        if 'token' in json_data.keys():
            token = Token.serialize_from_json(json_data['token'], response_type)
        
        return cls(card, token, json_response= json_data, response_type = response_type)
    
    @classmethod
    def create(cls, card: Card, token: Token) -> 'PaymentSource':
        return cls(card, token)


"""Referenced ReferencedPayouts resource client for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/referenced-ReferencedPayouts/v1/
"""

import json

from typing import Type, TypeVar, List
from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LEGACY_LIVE_API_BASE_URL,
    LEGACY_SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    ResponseType, 
    PaypalApiResponse, 
    PatchUpdateRequest,
    PaypalApiBulkResponse
)

from pypaypal.entities.rpayouts import (
    ExecutionType,
    ReferencedPayoutsItem,
    ReferencedPayoutRequest,
    ReferencedPayoutResponse
)

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LEGACY_LIVE_API_BASE_URL, 'payments')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(LEGACY_SANDBOX_API_BASE_URL, 'payments')

T = TypeVar('T', bound = 'ReferencedPayoutClient')

I = TypeVar('I', bound = 'ReferencedPayoutItemClient')

class ReferencedPayoutClient(ClientBase):
    """ReferencedPayout resource group client class.
    """
    
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)
    
    def create_referenced_batch_payout(
        self, request: ReferencedPayoutRequest, request_id: str = None, partnerAttrId: str = None) -> PaypalApiBulkResponse[ActionLink]:
        """Creates a referenced batch payout for asynchronous, offline processing
        
        Arguments:
            request {ReferencedPayoutRequest} -- Request processing info
        
        Keyword Arguments:
            request_id {str} -- paypal request id for idempotence (default: {None})
            partnerAttrId {str} -- paypal partner attribution id (default: {None})
        
        Returns:
            PaypalApiBulkResponse[ActionLink] -- Batch links
        """
        headers = dict()

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        if partnerAttrId:
            headers['PayPal-Partner-Attribution-Id'] = partnerAttrId
        if request.execution_type.is_async:
            headers['Prefer'] = 'respond-async'

        body = request.to_dict()
        body.pop('_execution_type', None)
        api_response = self._session.post(self._base_url, json.dumps(body), headers = headers)

        if api_response.status_code // 100 != 2:
            return PaypalApiBulkResponse.error(api_response)

        j_data = api_response.json()
        links = [ ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in j_data.get('links', []) ]

        return PaypalApiBulkResponse.success(api_response, links)
    
    def _process_referenced_list_response(self, api_response) -> PaypalApiResponse[ReferencedPayoutResponse]:
        """Process the API response for the list batch payouts calls
        
        Arguments:
            api_response {[type]} -- Http api response
        
        Returns:
            PaypalApiResponse[ReferencedPayoutResponse] -- response info
        """
        if api_response.status_code // 100 != 2:
            return PaypalApiBulkResponse.error(api_response)
        
        j_data = api_response.json()
        return PaypalApiBulkResponse.success(api_response, ReferencedPayoutResponse.serialize_from_json(j_data) if j_data else None)        

    def list_referenced_batch_payout_items(self, payouts_batch_id: str) -> PaypalApiResponse[ReferencedPayoutResponse]:
        """Calls the API to list the payout items in a referenced batch payout. Each item in the list includes payout item details. 
        
        Arguments:
            payouts_batch_id {str} -- The batch id
        
        Returns:
            PaypalApiResponse[ReferencedPayoutResponse] -- Listed referenced payouts with the directive
        """
        return self._process_referenced_list_response(self._session.get(parse_url(self._base_url, payouts_batch_id)))

    def list_referenced_batch_payout_items_by_link(self, link: ActionLink) -> PaypalApiResponse[ReferencedPayoutResponse]:
        """Calls the API to list the payout items in a referenced batch payout. Each item in the list includes payout item details. 
        
        Arguments:
            payouts_batch_id {str} -- The batch id
        
        Returns:
            PaypalApiResponse[ReferencedPayoutResponse] -- Listed referenced payouts with the directive
        """
        return self._process_referenced_list_response(self._session.get(link.href))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'referenced-payouts'), session)

class ReferencedPayoutItemClient(ClientBase):
    """Payout items resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def _process_item_response(self, api_response) -> PaypalApiResponse[ReferencedPayoutsItem]:
        """Process the api response for a referenced payout item 
        
        Arguments:
            api_response {[type]} -- The paypal API response
        
        Returns:
            PaypalApiResponse[ReferencedPayoutsItem] -- The processed item
        """
        if api_response.status_code // 100 != 2:
            return PaypalApiBulkResponse.error(api_response)

        j_data = api_response.json()        
        return PaypalApiBulkResponse.success(
            api_response, 
            ReferencedPayoutsItem.serialize_from_json(j_data) if j_data else None
        )

    def create_referenced_payout_item(
            self, item: ReferencedPayoutsItem, request_id: str = None, 
            partnerAttrId: str = None, execution_type: ExecutionType = ExecutionType.SYNC
        ) -> PaypalApiResponse[ReferencedPayoutsItem]:
        """Calls the API to create a referenced payout item.
        
        Arguments:
            item {ReferencedPayoutsItem} -- The item to be registered
        
        Keyword Arguments:
            request_id {str} -- request id for idempotence (default: {None})
            partnerAttrId {str} -- paypal partner attribution id (default: {None})
            execution_type {ExecutionType} -- how will the API process the transaction (default: {ExecutionType.SYNC})
        
        Returns:
            PaypalApiResponse[ReferencedPayoutsItem] -- Api Response obj with the item basic info (id & links).
        """
        headers = dict()

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        if partnerAttrId:
            headers['PayPal-Partner-Attribution-Id'] = partnerAttrId
        if execution_type.is_async:
            headers['Prefer'] = 'respond-async'

        return self._process_item_response(
            self._session.post(self._base_url, json.dumps(item.to_dict()), headers = headers)
        )
    
    def show_referenced_payout_item_details(self, payout_item_id: str, partnerAttrId: str = None) -> PaypalApiResponse[ReferencedPayoutsItem]:
        """Calls the paypal api to show a referenced payout item details.
        
        Arguments:
            payout_item_id {str} -- The item to be listed
        
        Keyword Arguments:
            partnerAttrId {str} -- paypal partner attribution id (default: {None})
        
        Returns:
            PaypalApiResponse[ReferencedPayoutsItem] -- Api Response obj with the item
        """
        url = parse_url(self._base_url, payout_item_id)
        headers = {'PayPal-Partner-Attribution-Id' : partnerAttrId } if payout_item_id else dict()
        return self._process_item_response(self._session.get(url, headers = headers))

    def show_referenced_payout_item_details_by_link(self, link: ActionLink, partnerAttrId: str = None) -> PaypalApiResponse[ReferencedPayoutsItem]:
        """Calls the paypal api to show a referenced payout item details.
        
        Arguments:
            link {ActionLink} -- Action link with the entity read url.8
        
        Keyword Arguments:
            partnerAttrId {str} -- paypal partner attribution id (default: {None})
        
        Returns:
            PaypalApiResponse[ReferencedPayoutsItem] -- Api Response obj with the item
        """
        url = link.href
        headers = {'PayPal-Partner-Attribution-Id' : partnerAttrId } if payout_item_id else dict()
        return self._process_item_response(self._session.get(url, headers = headers))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> I:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(parse_url(base_url, 'referenced-payouts-items'), session)


"""
    Payment resource refunds related entities
"""
from enum import Enum
from datetime import datetime
from typing import Type, List

import dateutil.parser

from pypaypal.entities.base import (
    T,
    Money,
    ActionLink,
    PlatformFee,
    ResponseType,    
    PayPalEntity
)

from pypaypal.entities.payments.commons import SellerProtection, ExchangeRate

class RefundStatusDetailReason(Enum):
    ECHECK = 1

class RefundStatus(Enum):
    PENDING = 1
    COMPLETED = 2
    CANCELED = 3

class RefundStatusDetail(PayPalEntity):
    """Refund status details
    """
    def __init__(self, reason: str, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.reason  = reason

    @property
    def reason_enum(self) -> RefundStatusDetailReason:
        try:
            return RefundStatusDetailReason[self.reason] if self.reason else None
        except:
            return None

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        return cls(json_data.get('reason'), json_response = json_data, response_type = response_type)

    @classmethod
    def create(cls, reason: RefundStatusDetailReason = RefundStatusDetailReason.ECHECK) -> 'RefundStatusDetail':
        return cls(reason.name)

class RefundNetAmtBreakdown(PayPalEntity):
    """net_amount_breakdown obj representation
    """
    def __init__(self, payable_amt: Money, converted_amt: Money, exchange_rate: ExchangeRate, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.payable_amount = payable_amt
        self.converted_amount = converted_amt
        self.exchange_rate = exchange_rate

    @classmethod    
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        payable_amount, converted_amount, exchange_rate = None, None, None

        if 'payable_amount' in json_data.keys():
            payable_amount = Money.serialize_from_json(json_data['payable_amount', response_type])
        if 'converted_amount' in json_data.keys():
            converted_amount = Money.serialize_from_json(json_data['converted_amount', response_type])
        if 'exchange_rate' in json_data.keys():
            exchange_rate = ExchangeRate.serialize_from_json(json_data['exchange_rate', response_type])
        
        return cls(
            payable_amount, converted_amount, exchange_rate, 
            json_response = json_data, response_type = response_type
        )

class SellerRefundBreakdown(PayPalEntity):
    """Capture seller receivable breakdown    
    """
    def __init__(
        self, gross_amt: Money, paypal_fee: Money, net_amt: Money, 
        total_refunded_amount: Money, net_amt_breakdown: RefundNetAmtBreakdown, 
        platform_fees: List[PlatformFee], **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))
        self.gross_amount = gross_amt
        self.paypal_fee = paypal_fee
        self.net_amount = net_amt
        self.total_refunded_amount = total_refunded_amount
        self.net_amount_breakdown = net_amt_breakdown
        self.platform_fees = platform_fees

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        platform_fees, net_amount_breakdown = [], None
        gross_amount, paypal_fee, net_amount, total_refunded_amount = None, None, None, None

        if 'gross_amount' in json_data.keys():
            gross_amount = Money.serialize_from_json(json_data['gross_amount'], response_type)
        if 'paypal_fee' in json_data.keys():
            paypal_fee = Money.serialize_from_json(json_data['paypal_fee'], response_type)
        if 'net_amount' in json_data.keys():
            net_amount = Money.serialize_from_json(json_data['net_amount'], response_type)
        if 'total_refunded_amount' in json_data.keys():
            total_refunded_amount = Money.serialize_from_json(json_data['total_refunded_amount'], response_type)
        if 'net_amount_breakdown' in json_data.keys():
            net_amount_breakdown = RefundNetAmtBreakdown.serialize_from_json(json_data['net_amount_breakdown'], response_type)

        if 'platform_fees' in json_data.keys():
            platform_fees = [PlatformFee.serialize_from_json(x, response_type) for x in json_data['platform_fees']]
        
        return cls(
            gross_amount, paypal_fee, net_amount, total_refunded_amount, 
            net_amount_breakdown, platform_fees, json_response = json_data, 
            response_type = response_type
        )

class Refund(PayPalEntity):
    """Refund obj representation.
    """

    def __init__(
        self, refund_id: str, status: str,
        invoice_id: str, amount: Money, note_to_payer: str, 
        status_details: RefundStatusDetail, seller_breakdown: SellerRefundBreakdown, **kwargs):
        super().__init__(kwargs.get('json_response', dict()), kwargs.get('response_type', ResponseType.MINIMAL))    
        self.id = refund_id
        self.status = status
        self.invoice_id = invoice_id
        self.amount = amount
        self.note_to_payer = note_to_payer
        self.status_details = status_details
        self.seller_breakdown = seller_breakdown
        self._create_time = self._json_response.get('create_time', kwargs.get('create_time'))
        self._update_time = self._json_response.get('update_time', kwargs.get('update_time'))
        self.links = [ActionLink(x['href'], x['rel'], x.get('method', 'GET')) for x in self._json_response.get('links', [])]

    @property
    def status_enum(self) -> RefundStatus:
        try:
            return RefundStatus[self.status] if self.status else None
        except:
            return None

    @property
    def create_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._create_time) if self._create_time else None
        except:
            return None

    @property
    def update_time(self) -> datetime:
        try:
            return dateutil.parser.parse(self._update_time) if self._update_time else None
        except:
            return None

    @property
    def Status_enum(self) -> RefundStatusDetailReason:
        try:
            return RefundStatusDetailReason[self.status] if self.status else None
        except:
            return None

    @property
    def read_link(self) -> ActionLink:
        """Retrieves a link to read this entity details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'self', self.links), None)

    @property
    def up_link(self) -> ActionLink:
        """Retrieves a link to read this entity capture details.
        
        Returns:
            ActionLink -- The link for requesting the information to the API.
        """
        return next(filter(lambda x: x.rel == 'up', self.links), None)

    @classmethod
    def serialize_from_json(cls: Type[T], json_data: dict, response_type: ResponseType = ResponseType.MINIMAL) -> T:
        amount, status_details, seller_breakdown = None, None, None
        
        if 'amount' in json_data.keys():
            amount = Money.serialize_from_json(json_data['amount'], response_type)
        if 'status_details' in json_data.keys():
            status_details = RefundStatusDetail.serialize_from_json(json_data['status_details'], response_type)
        if 'seller_breakdown' in json_data.keys():
            seller_breakdown = SellerRefundBreakdown.serialize_from_json(json_data['seller_breakdown'], response_type)

        return cls(
            json_data.get('id'), json_data.get('status'), json_data.get('invoice_id'), 
            amount, json_data.get('note_to_payer'), status_details, seller_breakdown,
            json_response = json_data, response_type = response_type
        )



"""
    Payment captures resource clients for the paypal resource group.

    Resource docs & Reference: https://developer.paypal.com/docs/api/payments/v2/ 
"""
import json

from datetime import datetime
from typing import Type, TypeVar, List

from pypaypal.clients.base import ClientBase, ActionLink

from pypaypal.http import ( 
    parse_url,
    PayPalSession,
    LIVE_API_BASE_URL,
    SANDBOX_API_BASE_URL
)

from pypaypal.entities.base import ( 
    Money,
    ResponseType,
    PaypalApiResponse
)

from pypaypal.entities.payments.refunds import Refund
from pypaypal.entities.payments.captures import Capture

"""
    Base Resource Live URL
"""
_LIVE_RESOURCE_BASE_URL = parse_url(LIVE_API_BASE_URL, 'payments', 'captures')

"""
    Base Resource Sandbox URL
"""
_SANDBOX_RESOURCE_BASE_URL = parse_url(SANDBOX_API_BASE_URL, 'payments', 'captures')

T = TypeVar('T', bound = 'CaptureClient')

class CaptureClient(ClientBase):
    """Captures resource group client class.
    """
    def __init__(self, url: str, session: PayPalSession):
        super().__init__(url, session)

    def show_capture_details(self, capture_id: str) -> PaypalApiResponse[Capture]:
        """Calls the paypal API to show details for an authorized payment, by ID.
        
        Arguments:
            capture_id {str} -- The capture id
        
        Returns:
            PaypalApiResponse[Capture] -- An api response with the capture details
        """
        api_response = self._session.get(parse_url(self._base_url, capture_id))

        if api_response.status_code != 200:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Capture.serialize_from_json(api_response.json()))

    def refund_capture(
        self, capture_id: str, invoice_id: str, note_to_payer: str, 
        amount: Money = None, request_id: str = None, auth_assertion_token: str = None,
        response_type: ResponseType = ResponseType.MINIMAL) -> PaypalApiResponse[Refund]:
        """Calls the api to refund a capture
        
        Arguments:
            capture_id {str} -- capture identifier
            invoice_id {str} -- invoice related capture
            note_to_payer {str} -- notes to the customer
        
        Keyword Arguments:
            amount {Money} -- amount to be refunded if None then 'captured amount - previous refunds' (default: {None})
            request_id {str} -- request id for idempotence (default: {None})
            auth_assertion_token {str} -- auth assertion token. See paypal header docs (default: {None})
            response_type {ResponseType} -- response type. See paypal header docs (default: {ResponseType.MINIMAL})
        
        Returns:
            PaypalApiResponse[Refund] -- api response with refund details
        """

        body = dict()
        url = parse_url(self._base_url, capture_id, 'refund')

        headers = {
            'Prefer': response_type.as_header_value()            
        }

        if request_id:
            headers['PayPal-Request-Id'] = request_id
        if auth_assertion_token:
            headers['PayPal-Auth-Assertion'] = auth_assertion_token

        if invoice_id:
            body['invoice_id'] = invoice_id
        if note_to_payer:
            body['note_to_payer'] = note_to_payer
        if amount:
            body['amount'] = amount.to_dict()

        api_response = self._session.post(url, json.dumps(body), headers = headers)

        if api_response.status_code != 201:
            return PaypalApiResponse(True, api_response)

        return PaypalApiResponse(False, api_response, Refund.serialize_from_json(api_response.json()))

    @classmethod
    def for_session(cls: T, session: PayPalSession) -> T:
        """Creates a client from a given paypal session
        
        Arguments:
            cls {T} -- class reference
            session {PayPalSession} -- the paypal session
        
        Returns:
            T -- an instance of Dispute client with the right configuration by session mode
        """
        base_url = _LIVE_RESOURCE_BASE_URL if session.session_mode.is_live() else _SANDBOX_RESOURCE_BASE_URL
        return cls(base_url, session)
